import json
from transpiler.utils import cairoize_bytes
from web3 import Web3
from typing import Any, Dict, Optional, Union
import aiohttp
import asyncio
from starkware.starknet.services.api.gateway.transaction import (
    InvokeFunction,
    Transaction,
)
from transpiler.EvmToCairo import EvmToCairo, parse_operations
from eth_hash.auto import keccak
import os
from starkware.cairo.lang.compiler.parser import parse_file

TEST_ENV = ""
artifacts_dir = os.path.join(os.path.expanduser("~"), ".warp", "artifacts")


def get_selector_cairo(args: str) -> int:
    return int.from_bytes(keccak(args.encode("ascii")), "big") & (2 ** 250 - 1)


async def send_req(method, url, tx: Optional[Union[str, Dict[str, Any]]] = None):
    if tx is not None:
        async with aiohttp.ClientSession() as session:
            async with session.request(
                method=method, url=url, data=Transaction.Schema().dumps(obj=tx)
            ) as response:
                text = await response.text()
                return text
    else:
        async with aiohttp.ClientSession() as session:
            async with session.request(method=method, url=url, data=None) as response:
                text = await response.text()
                return text


def _transpile(file):
    evm_to_cairo = EvmToCairo(cur_evm_pc=0)
    with open(file, "r") as opcodes_file:
        operations = list(parse_operations(opcodes_file))
        for op in operations:
            op.inspect_program(operations)

        for op in operations:
            evm_to_cairo.process_operation(op)

    os.remove(file)
    return parse_file(evm_to_cairo.finish(True)).format()


# returns true/false on transaction success/failure
async def _invoke(source_name, address, cairo_abi, function, inputs):
    with open(os.path.join(artifacts_dir, "selector_jumpdests.json")) as f:
        jumpdests = json.load(f)
    with open(cairo_abi) as f:
        cairo_abi = json.load(f)
    with open(os.path.join(artifacts_dir, f"{source_name[:-4]}_abi.json")) as f:
        evm_abi = json.load(f)
    with open(
        os.path.join(artifacts_dir, f"{source_name[:-4]}_bytecode"),
        "r",
    ) as f:
        bytecode = f.read()

    w3 = Web3()
    evm_contract = w3.eth.contract(abi=evm_abi, bytecode=bytecode)
    evm_calldata = evm_contract.encodeABI(fn_name=function, args=inputs)
    cairo_input, unused_bytes = cairoize_bytes(bytes.fromhex(evm_calldata[2:]))
    calldata_size = (len(cairo_input) * 16) - unused_bytes
    try:
        address = int(address, 16)
    except ValueError:
        raise ValueError("Invalid address format.")

    selector = get_selector_cairo("main")
    calldata = [calldata_size, unused_bytes, len(cairo_input)] + cairo_input
    print(f"calldata: {calldata}")
    tx = InvokeFunction(
        contract_address=address, entry_point_selector=selector, calldata=calldata
    )

    response = await send_req(
        method="POST", url=f"{TEST_ENV}/gateway/add_transaction", tx=tx
    )
    tx_id = json.loads(response)["tx_id"]
    print(
        f"""\
Invoke transaction was sent.
Contract address: 0x{address:064x}.
Transaction ID: {tx_id}."""
    )
    return True


async def _call(address, abi, function, inputs) -> bool:
    with open(abi) as f:
        abi = json.load(f)

    try:
        address = int(address, 16)
    except ValueError:
        raise ValueError("Invalid address format.")

    selector = get_selector_cairo(function)
    calldata = inputs
    tx = InvokeFunction(
        contract_address=address, entry_point_selector=selector, calldata=calldata
    )

    url = f"{TEST_ENV}/feeder_gateway/call_contract?blockId=null"
    async with aiohttp.ClientSession() as session:
        async with session.request(method="POST", url=url, data=tx.dumps()) as response:
            raw_resp = await response.text()
            resp = json.loads(raw_resp)
        return resp["result"][0]



import operator

from transpiler.Imports import UINT256_MODULE
from transpiler.Operations.Binary import Binary, SimpleBinary
from transpiler.Operations.Ternary import Ternary
from transpiler.utils import (
    uint256_to_int256,
    int256_to_uint256,
    UINT256_HALF_BOUND,
    is_bit_set,
    bit_not,
)


class Add(SimpleBinary):
    def __init__(self):
        super().__init__(operator.add, UINT256_MODULE, "uint256_add")

    def generate_cairo_code(self, op1, op2, res):
        return [f"let (local {res} : Uint256, _) = {self.function_name}({op1}, {op2})"]


class Mul(SimpleBinary):
    def __init__(self):
        super().__init__(operator.mul, UINT256_MODULE, "uint256_mul")

    def generate_cairo_code(self, op1, op2, res):
        return [f"let (local {res} : Uint256, _) = {self.function_name}({op1}, {op2})"]


class Sub(SimpleBinary):
    def __init__(self):
        super().__init__(operator.sub, UINT256_MODULE, "uint256_sub")


class Div(Binary):
    def evaluate_eagerly(self, a, b):
        return a // b if b != 0 else 0

    def generate_cairo_code(self, op1, op2, res):
        return [
            f"let (local {res} : Uint256, _) = uint256_unsigned_div_rem({op1}, {op2})"
        ]

    @classmethod
    def required_imports(cls):
        return {UINT256_MODULE: {"uint256_unsigned_div_rem"}}


class Sdiv(Binary):
    def evaluate_eagerly(self, a, b):
        a = uint256_to_int256(a)
        b = uint256_to_int256(b)
        if a == -UINT256_HALF_BOUND and b == -1:
            return a
        return int256_to_uint256(a // b) if b != 0 else 0

    def generate_cairo_code(self, op1, op2, res):
        return [
            f"let (local {res} : Uint256, _) = uint256_signed_div_rem({op1}, {op2})"
        ]

    @classmethod
    def required_imports(cls):
        return {UINT256_MODULE: {"uint256_signed_div_rem"}}


class Exp(SimpleBinary):
    def __init__(self):
        # FIXME we should probably use a fast modular exponentiation algorithm instead
        super().__init__(operator.pow, UINT256_MODULE, "uint256_exp")


def mod(a, b):
    return a % b if b != 0 else 0


class Mod(SimpleBinary):
    def __init__(self):
        super().__init__(mod, UINT256_MODULE, "uint256_mod")


def smod(a, b):
    a = uint256_to_int256(a)
    b = uint256_to_int256(b)
    return int256_to_uint256(a % b) if b != 0 else 0


class SMod(SimpleBinary):
    def __init__(self):
        super().__init__(smod, "evm.uint256", "smod")


class AddMod(Ternary):
    def evaluate_eagerly(self, a, b, c):
        return (a + b) % c

    def generate_cairo_code(self, op1, op2, op3, res):
        return [f"let (local {res} : Uint256) = uint256_addmod({op1}, {op2}, {op3})"]

    @classmethod
    def required_imports(cls):
        return {UINT256_MODULE: {"uint256_addmod"}}


class MulMod(Ternary):
    def evaluate_eagerly(self, a, b, c):
        return a * b % c

    def generate_cairo_code(self, op1, op2, op3, res):
        return [f"let (local {res} : Uint256) = uint256_mulmod({op1}, {op2}, {op3})"]

    @classmethod
    def required_imports(cls):
        return {UINT256_MODULE: {"uint256_mulmod"}}


class SignExtend(Binary):
    def evaluate_eagerly(self, b, x):
        if b > 31: return x
        bit = (b * 8) + 7
        mask = (1 << bit) - 1
        if is_bit_set(x, bit):
            return x | bit_not(mask)
        else:
            return x & mask

    def generate_cairo_code(self, op1, op2, res):
        # notice, that op1 is the byte position, so it should be the
        # second arg to uint256_signextend
        return [f"let (local {res} : Uint256) = uint256_signextend({op2}, {op1})"]

    @classmethod
    def required_imports(cls):
        return {UINT256_MODULE: {"uint256_signextend"}}



import operator

from transpiler.Operations.Binary import Binary, SimpleBinary
from transpiler.Operations.Unary import Unary
from transpiler.utils import uint256_to_int256


class IsZero(Unary):
    def evaluate_eagerly(self, x):
        return x == 0

    def generate_cairo_code(self, op, res):
        return [
            f"let (local {res} : Uint256) = is_zero{{range_check_ptr=range_check_ptr}}({op})"
        ]

    @classmethod
    def required_imports(cls):
        return {"evm.uint256": {"is_zero"}}


class Eq(Binary):
    @classmethod
    def evaluate_eagerly(self, x, y):
        return x == y

    def generate_cairo_code(self, op1, op2, res):
        return [
            f"let (local {res} : Uint256) = is_eq{{range_check_ptr=range_check_ptr}}({op1}, {op2})"
        ]

    @classmethod
    def required_imports(cls):
        return {"evm.uint256": {"is_eq"}}


class Lt(Binary):
    def evaluate_eagerly(self, x, y):
        return x < y

    def generate_cairo_code(self, op1, op2, res):
        return [
            "local memory_dict : DictAccess* = memory_dict",
            f"let (local {res} : Uint256) = is_lt{{range_check_ptr=range_check_ptr}}({op1}, {op2})",
        ]

    @classmethod
    def required_imports(cls):
        return {"evm.uint256": {"is_lt"}}


class Gt(Binary):
    def evaluate_eagerly(self, x, y):
        return x > y

    def generate_cairo_code(self, op1, op2, res):
        return [
            f"let (local {res} : Uint256) = is_gt{{range_check_ptr=range_check_ptr}}({op1}, {op2})"
        ]

    @classmethod
    def required_imports(cls):
        return {"evm.uint256": {"is_gt"}}


def slt(a, b):
    return uint256_to_int256(a) < uint256_to_int256(b)


class Slt(Binary):
    def evaluate_eagerly(self, x, y):
        return slt(x, y)

    def generate_cairo_code(self, op1, op2, res):
        return [
            f"let (local {res} : Uint256) = slt{{range_check_ptr=range_check_ptr}}({op1}, {op2})"
        ]

    @classmethod
    def required_imports(cls):
        return {"evm.uint256": {"slt"}}


def sgt(a, b):
    return uint256_to_int256(a) > uint256_to_int256(b)


class Sgt(Binary):
    def evaluate_eagerly(self, x, y):
        return sgt(x, y)

    def generate_cairo_code(self, op1, op2, res):
        return [
            f"let (local {res} : Uint256) = sgt{{range_check_ptr=range_check_ptr}}({op1}, {op2})"
        ]

    @classmethod
    def required_imports(cls):
        return {"evm.uint256": {"sgt"}}



import os
import asyncio
import click
from enum import Enum
from cli.compilation.Compile import Vyper, Solidity
from cli.commands import _transpile, _call, _invoke


class Command(Enum):
    INVOKE = 0
    CALL = 1


@click.group()
def warp():
    pass


@warp.command()
@click.argument("contract_path", nargs=1, type=click.Path(exists=True))
def transpile(contract_path):
    if contract_path.endswith("vy"):
        path = os.path.abspath(click.format_filename(contract_path))
        contract = Vyper(path)
        cairo_str = _transpile(contract.opcodes)
        with open(f"{path[:-3]}.cairo", "w") as f:
            f.write(cairo_str)
    elif contract_path.endswith("sol"):
        path = os.path.abspath(click.format_filename(contract_path))
        filename = os.path.basename(path)
        contract = Solidity(path)
        with open(f"{path[:-4]}.opcode", "w") as f:
            f.write(contract.opcodes_str)
        cairo_str = _transpile(
            f"{path[:-4]}.opcode",
        )
        with open(f"{path[:-4]}.cairo", "w") as f:
            f.write(cairo_str)
    return None


return_args = {}


@warp.command()
@click.option("--contract", required=True)
@click.option("--address", required=True)
@click.option("--abi", required=True, type=click.Path(exists=True))
@click.option("--function", required=True)
@click.option("--inputs", required=True)
def invoke(contract, address, abi, function, inputs):
    inputs = inputs.split(" ")
    for idx, input in enumerate(inputs):
        if input.isdigit():
            inputs[idx] = int(input)
    return_args["address"] = address
    return_args["abi"] = abi
    return_args["function"] = function
    return_args["contract"] = contract
    return_args["inputs"] = inputs
    return_args["type"] = Command.INVOKE


@warp.command()
@click.option("--address", required=True, type=click.Path(exists=True))
@click.option("--abi", required=True, type=click.Path(exists=True))
@click.option("--function", required=True, type=click.Path(exists=True))
def call(address, abi, function):
    _call(address, abi, function)


cli = click.CommandCollection(sources=[warp])


async def test():
    await asyncio.sleep(1)
    print("DONE!")


def main():
    try:
        warp()
    except:
        if return_args != {}:
            if return_args["type"] is Command.INVOKE:
                asyncio.run(
                    _invoke(
                        return_args["contract"],
                        return_args["address"],
                        return_args["abi"],
                        return_args["function"],
                        return_args["inputs"],
                    )
                )



from __future__ import annotations
from typing import Union

import transpiler.StackValue as StackValue


class EvmStack:
    """
    Simulates the EVM stack.
    """

    def __init__(self):
        # The last value in self.values represents the value at the top of the stack.
        self.values: list[StackValue.StackValue] = []
        # If more values of the stack are required, those values will be dynamically added.
        # self.depth represents how many such values were added to self.values.
        self.depth = 0

    def pad_up_to(self, n):
        """
        Makes sure there are at least n elements in the stack. Adds variables to the bottom
        of the stack if needed.
        """
        padding = max(0, n - len(self.values))
        new_depth = self.depth + padding
        padding_values = [f"stack{d}.value" for d in range(self.depth, new_depth)]
        padding_values.reverse()

        self.depth = new_depth
        self.values = padding_values + self.values

    def push_ref(self, x: str):
        self.values.append(StackValue.Str(x))

    def push_uint256(self, x: Union[int, StackValue.Uint256]):
        self.values.append(
            x if isinstance(x, StackValue.Uint256) else StackValue.Uint256(x)
        )

    def pop(self):
        self.pad_up_to(1)
        return self.values.pop(-1)

    def dup(self, i):
        v = self.get(i)
        self.values.append(v)

    def get(self, i):
        self.pad_up_to(i)
        return self.values[-i]

    def swap(self, i):
        self.pad_up_to(i + 1)
        self.values[-i - 1], self.values[-1] = self.values[-1], self.values[-i - 1]

    def __repr__(self):
        return repr(self.values)

    def prepare(self):
        """
        Creates local variables for the stack items that were read.
        """
        return [
            f"local stack{i + 1} : StackItem* = stack{i}.next"
            for i in range(self.depth)
        ]

    def build_stack_instructions(self):
        """
        Returns the Cairo instruction for constructing the stack.
        """
        for i, val in enumerate(self.values):
            if val != f"stack{self.depth - i - 1}.value":
                break
        else:
            i = len(self.values)
        cur_stack = f"stack{self.depth - i}"

        res = []
        for i, val in enumerate(self.values[i:], start=i):
            res.append(
                f"local newitem{i} : StackItem = "
                f"StackItem(value={val}, next={cur_stack})"
            )
            cur_stack = f"&newitem{i}"
        return res, cur_stack



from transpiler.Operations.EnforcedStack import EnforcedStack

from transpiler.Operation import Operation
from transpiler.StackValue import Uint256

class Unary(EnforcedStack):
    def __init__(self):
        super().__init__(n_args=1, has_output=True)



from transpiler.Operation import Operation
from transpiler.StackValue import Uint256

EMPTY_OUTPUT = "Output(1, cast(0, felt*), 0)"


class Stop(Operation):
    def proceed(self, state):
        state.unreachable = True
        return state.make_return_instructions(Uint256(0), EMPTY_OUTPUT)


EMPTY_OUTPUT = "Output(1, cast(0, felt*), 0)"

"""
In the future, when StarkWare implements
revert semantics in the StarkNet VM, this will change
to something like:
return [f"[ap -1] = get_memory({offset}, {offset+size})", "assert 0 = 1"]
"""


class Revert(Operation):
    def proceed(self, state):
        offset = state.stack.pop()
        size = state.stack.pop()
        state.unreachable = True
        return [
            "assert 0 = 1",
            "local item : StackItem =StackItem(value=Uint256(0, 0),next=stack0)",
            "return (stack=&item, evm_pc=Uint256(-1, 0), output=Output(0, cast(0, felt*), 0))",
        ]


class Invalid(Operation):
    def proceed(self, state):
        state.unreachable = True
        return [
            "local item : StackItem =StackItem(value=Uint256(0, 0),next=stack0)",
            "assert 0 = 1",
            "return (stack=&item,evm_pc=Uint256(-1, 0), output=Output(0, cast(0, felt*), 0))",
        ]


class Return(Operation):
    def proceed(self, state):
        state.unreachable = True
        offset = state.stack.pop().get_low_bits()
        length = state.stack.pop().get_low_bits()
        return [
            f"let (local output : Output) = create_from_memory({offset}, {length})",
            *state.make_return_instructions(Uint256(0), "output"),
        ]

    def required_imports(self):
        return {"evm.output": {"create_from_memory"}}



from transpiler.Imports import UINT256_MODULE
from transpiler.Operation import Operation


class Jumpdest(Operation):
    def proceed(self, state):
        return state.make_return_instructions(state.cur_evm_pc)

    def process_structural_changes(self, evmToCairo):
        evmToCairo.finish_segment()
        evmToCairo.start_new_segment()


class Jump(Operation):
    def proceed(self, state):
        state.unreachable = True
        a = state.stack.pop()
        return state.make_return_instructions(a)


class JumpI(Operation):
    def proceed(self, state):
        a = state.stack.pop()
        b = state.stack.pop()
        return_instructions = state.make_return_instructions(a)
        return [
            f"let (immediate) = uint256_eq{{range_check_ptr=range_check_ptr}}({b}, Uint256(0, 0))",
            "if immediate == 0:",
            *return_instructions,
            "end",
        ]

    def required_imports(self):
        return {UINT256_MODULE: {"uint256_eq"}}



from __future__ import annotations
from transpiler.StackValue import Uint256, Str


def is_valid_uintN(n: int, x: int):
    return 0 <= x < 2 ** n


UINT256_BOUND = 2 ** 256
UINT256_HALF_BOUND = 2 ** 255
UINT128_BOUND = 2 ** 128


def get_low_high(x):
    if isinstance(x, str):
        return f"{x}.low", f"{x}.high"
    elif isinstance(x, Str):
        return f"{x}.low", f"{x}.high"
    elif isinstance(x, Uint256):
        return x.get_low_high()


def int256_to_uint256(x: int) -> int:
    assert -UINT256_HALF_BOUND <= x < UINT256_HALF_BOUND
    return UINT256_BOUND + x if x < 0 else x


def uint256_to_int256(x: int) -> int:
    assert 0 <= x < UINT256_BOUND
    return x - UINT256_BOUND if x >= UINT256_HALF_BOUND else x


def cairoize_bytes(bs: bytes) -> tuple(list[int], int):
    """Represent bytes as an array of 128-bit big-endian integers and
    return a number of unused bytes in the last array cell.
    """
    unused_bytes = -len(bs) % 16
    bs = bs.ljust(len(bs) + unused_bytes, b"\x00")
    arr = [int.from_bytes(bs[i : i + 16], "big") for i in range(0, len(bs), 16)]
    return (arr, unused_bytes)


# is_bit_set returns true if bit n-th is set, where n = 0 is LSB.
# The n must be <= 255.
def is_bit_set(x, n):
    x_bin_str = bin(x)[2:]
    x_bin_str = "0b" + "0" * (256 - len(x_bin_str)) + x_bin_str
    return int(x_bin_str[-n]) != 0


def bit_not(n, numbits=256):
    return (1 << numbits) - 1 - n


from __future__ import annotations
from transpiler.Operation import Operation

class CodeSize(Operation):
    code_size: int

    @classmethod
    def required_imports(cls):
        return {}

    def proceed(self, state):
        state.stack.push_uint256(self.code_size)
        return []

    @classmethod
    def inspect_program(self, code: list[Operation]):
        code_size = 0
        for op in code:
            code_size += op.size_in_bytes()
        self.code_size = code_size



import abc

from transpiler.Operation import Operation
from transpiler.StackValue import UINT256_BOUND, Uint256, Str
from transpiler.Imports import Imports

# Important: per yellow paper, the representation of uint256 as bytes
# is big-endian.


class MemoryAccess(Operation):
    def proceed(self, state):
        try:
            address = state.stack.pop().get_low_bits()
        except:
            address = f"{state.stack.pop()}.low"
        access_operations = self._do_memory_access(address, state)
        return [
            f"let (local msize) = update_msize{{range_check_ptr=range_check_ptr}}(msize, {address}, {self.access_width()})",
            "local memory_dict : DictAccess* = memory_dict",
            "local msize = msize",
            *access_operations,
        ]

    def required_imports(self):
        return {"evm.utils": {"update_msize"}, **self._memory_access_imports()}

    @classmethod
    @abc.abstractmethod
    def access_width(cls) -> int:
        pass

    @abc.abstractmethod
    def _do_memory_access(self, address, state):
        pass

    @classmethod
    @abc.abstractmethod
    def _memory_access_imports(cls) -> Imports:
        pass


class MStore(MemoryAccess):
    @classmethod
    def access_width(cls):
        return 32

    def _do_memory_access(self, address, state):
        value = state.stack.pop()
        return [
            f"mstore{{memory_dict=memory_dict, range_check_ptr=range_check_ptr}}(offset={address}, value={value})",
            "local memory_dict : DictAccess* = memory_dict",
        ]

    @classmethod
    def _memory_access_imports(cls):
        return {
            "evm.memory": {"mstore"},
            "starkware.cairo.common.dict_access": {"DictAccess"},
        }


class MStore8(MemoryAccess):
    @classmethod
    def access_width(cls):
        return 1

    def _do_memory_access(self, address, state):
        value = state.stack.pop()
        return [
            f"let (local byte, _) = extract_lowest_byte({value})",
            f"mstore8(offset={address}, byte=byte)",
            "local memory_dict : DictAccess* = memory_dict",
        ]

    @classmethod
    def _memory_access_imports(cls):
        return {
            "evm.memory": {"mstore8"},
            "evm.uint256": {"extract_lowest_byte"},
        }


class MLoad(MemoryAccess):
    @classmethod
    def access_width(cls):
        return 32

    def _do_memory_access(self, address, state):
        res_ref_name = state.request_fresh_name()
        state.stack.push_ref(res_ref_name)
        return [
            f"let (local {res_ref_name} : Uint256) = mload{{memory_dict=memory_dict, range_check_ptr=range_check_ptr}}({address})",
            "local memory_dict : DictAccess* = memory_dict",
        ]

    @classmethod
    def _memory_access_imports(cls):
        return {"evm.memory": {"mload"}}


class MSize(Operation):
    def proceed(self, state):
        res_ref_name = state.request_fresh_name()
        state.stack.push_ref(res_ref_name)
        # the yellow paper defines msize as the memory size in bytes,
        # but rounded to the closest greater word boundary, i.e. a
        # 32-multiple. Also, it's most probably < 2**64.
        return [
            "let (local immediate) = round_up_to_multiple(msize, 32)",
            f"local {res_ref_name} : Uint256 = Uint256(immediate, 0)",
        ]

    def required_imports(self):
        return {"evm.utils": {"round_up_to_multiple", "get_max"}}



from __future__ import annotations
import os, sys
from typing import Dict, List, Tuple

WARP_ROOT = os.path.abspath(os.path.join(__file__, "../../../.."))
sys.path.append(os.path.join(WARP_ROOT, "src"))
from eth_hash.auto import keccak
from cli.compilation.Contract import Contract, Language
import json


def get_contract_lang(file_path) -> Language:
    if file_path.endswith("vy"):
        return Language.VYPER
    elif file_path.endswith("sol"):
        return Language.SOL
    else:
        raise Exception(
            "The language you are trying to transpile is not supported... YET"
        )


def read_file_json(file_name: str) -> List[str]:
    with open(file_name) as f:
        data = f.readlines()
    abi = []
    for line in data:
        if not "[" in line:
            continue
        else:
            abi.extend(json.loads(line))
    return abi


def get_func_sigs(abi: List[Dict[str, str]]) -> Dict[str, int]:
    sigs = {}
    for item in abi:
        if item["type"] != "function":
            continue
        else:
            name = item["name"] + "("
            if len(item["inputs"]) == 0:
                name += ")"
                if item["stateMutability"] == "payable":
                    sigs[name] = 1
                else:
                    sigs[name] = 0
                continue
            for idx, x in enumerate(item["inputs"]):
                if (idx == len(item["inputs"]) - 1) or item["inputs"] == "":
                    name += x["type"] + ")"
                else:
                    name += x["type"] + ","
            if item["stateMutability"] == "payable":
                sigs[name] = 1
            else:
                sigs[name] = 0
    return sigs


# CALLVALUE
# ISZERO
# PUSH2
# 0x00b3
# JUMPI
# PUSH1
# 0x00
# DUP1
# REVERT
# JUMPDEST
def is_payable_check_seq(opcodes: List[str], language: Language):
    if language is Language.SOL:
        is_payable_seq = (
            opcodes[0] == "CALLVALUE"
            and opcodes[1] == "ISZERO"
            and "PUSH" in opcodes[2]
            and opcodes[4] == "JUMPI"
            and "PUSH" in opcodes[5]
            and opcodes[7] == "DUP1"
            and opcodes[8] == "REVERT"
            and opcodes[9] == "JUMPDEST"
        )
        replace_seq = [
            "PASS",
            "PASS",
            "NOOP" + opcodes[2][-1],
            opcodes[3],
            "PASS",
            "NOOP" + opcodes[5][-1],
            opcodes[6],
            "PASS",
            "PASS",
            "PASS",
        ]
        return is_payable_seq, replace_seq
    return False


# DUP1
# PUSH4
# 0x6FDDE03
# EQ
# PUSH2
# 0xEA
# JUMPI
def is_entry_seq(opcodes: List[str], language: Language) -> bool:
    if language is Language.SOL:
        if opcodes[0] != "DUP1":
            return False
        else:
            return (
                opcodes[0] == "DUP1"
                and (opcodes[1] == "PUSH4" or opcodes[1] == "PUSH3")
                and opcodes[3] == "EQ"
                and "PUSH" in opcodes[4]
                and opcodes[6] == "JUMPI"
            )
    elif language is Language.VYPER:
        if opcodes[0] != "PUSH4":
            return False
        else:
            return (
                opcodes[0] == "PUSH4"
                and opcodes[2] == "DUP2"
                and opcodes[3] == "EQ"
                and opcodes[4] == "ISZERO"
                and "PUSH" in opcodes[5]
                and opcodes[7] == "JUMPI"
            )


def get_selectors(abi: List[Dict[str, str]]) -> Dict[str, str]:
    file_name = os.path.join(
        os.path.expanduser("~"), ".warp", "artifacts", "selectors.json"
    )
    sigs = get_func_sigs(abi)
    selectors = {}
    for sig in sigs.keys():
        selector = int("0x" + keccak(sig.encode("ascii")).hex()[:8], 16)
        selectors[selector] = {
            "signature": sig,
            "payable": sigs[sig],
        }
    with open(file_name, "w") as f:
        json.dump(selectors, f, indent=4)
    return selectors


def get_jumpdest_offset(language):
    if language is Language.VYPER:
        increment = 8
        jumpdest_pos_offset = 6
        return jumpdest_pos_offset, increment
    elif language is Language.SOL:
        increment = 7
        jumpdest_pos_offset = 5
        return jumpdest_pos_offset, increment
    else:
        print(language)


def get_selector_jumpdests(contract: Contract) -> Dict[str, Dict[str, int]]:
    file_name = os.path.join(
        os.path.expanduser("~"), ".warp", "artifacts", "selector_jumpdests.json"
    )
    selector_jumpdests = {}
    selectors = list(contract.selectors.keys())
    jumpdest_pos_offset, increment = get_jumpdest_offset(contract.lang)
    idx = 0
    while len(selectors) > 0:
        if idx + increment >= len(contract.opcodes):
            break
        seq = contract.opcodes[idx : idx + increment]
        is_entry = is_entry_seq(seq, contract.lang)
        if is_entry:
            try:
                selector = int(contract.opcodes[idx + 2], 16)
                func_sig = contract.selectors[selector]["signature"]
                selector_jumpdests[selector] = {
                    "signature": func_sig,
                    "payable": contract.selectors[selector]["payable"],
                    "jumpdest": contract.opcodes[idx + jumpdest_pos_offset],
                }
                selectors.remove(selector)
                idx += increment
                continue
            except KeyError:
                # for some reason solidity 4.18 can generate 3-byte function selectors
                print("Inside key error")
                selector = contract.opcodes[idx + 2].lower()
                for sel in selectors:
                    if int(selector, 16) == int(sel, 16):
                        print(sel, selector)
                idx += 1
                continue
        idx += 1
    with open(file_name, "w") as f:
        json.dump(selector_jumpdests, f, indent=4)
    return selector_jumpdests



from __future__ import annotations
from typing import Optional
import abc

from transpiler.StackValue import Uint256, UINT256_BOUND
from transpiler.Operation import Operation


def get_arg(state, arg_spec):
    assert arg_spec == "l" or arg_spec == "w"
    arg = state.stack.pop()
    return arg.get_low_bits() if arg_spec == "l" else arg


class EnforcedStack(Operation):
    def __init__(
        self, args_spec: str = None, n_args: int = None, has_output: bool = False
    ):
        """Parameters:

        'args_spec' — specifies arguments number and type. It is a
          string of argument types. Each type is either "w" — *w*hole
          Uint256, or "l" — just *l*ower 128-bits. It is useful since
          many EVM values are expected to be less than 2^64 and
          working with simple felts is more efficient in cairo. By
          default, if 'n_args' is specified, it is assumed to be a
          string of "w"s, i.e. all arguments are whole Uin256's.

        'n_args' — a number of arguments. By default, can be inferred
          from 'args_spec' if it is specified.

        'has_output' — whether the operation has output to put on top
          of the EVM stack.

        """
        if args_spec is None:
            assert (
                n_args is not None
            ), "Either supply a number of arguments, or their spec"
            self.args_spec = "w" * n_args
        else:
            assert (n_args is None) or n_args == len(
                args_spec
            ), "Arguments number must conform to their spec"
            self.args_spec = args_spec
        self.has_output = has_output

    def __try_eager_evaluation(self, state, args: list[Uint256]):
        try:
            output = self.evaluate_eagerly(*(arg.x for arg in args))
        except NotImplementedError:
            return False
        assert self.has_output == (output is not None)
        if output is not None:
            state.stack.push_uint256(output % UINT256_BOUND)
        return True

    def proceed(self, state: "EvmToCairo.SegmentState") -> list[str]:
        args = [get_arg(state, arg_spec) for arg_spec in self.args_spec]
        if all(isinstance(arg, Uint256) for arg in args):
            if self.__try_eager_evaluation(state, args):
                return []
        if self.has_output:
            output_name = state.request_fresh_name()
            state.stack.push_ref(output_name)
            return self.generate_cairo_code(*args, output_name)
        else:
            return self.generate_cairo_code(*args)

    def evaluate_eagerly(self, *args: tuple[int]) -> Optional[int]:
        """Eagerly evaluates the operation, given 'self.n_args' uint256
        arguments starting from the top of the stack.

        If there is only one output, you can return it without
        wrapping into a tuple. If eager evaluation is not possible (or
        feasible), return 'None'.

        Note: all outputs are used modulo 2^256.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def generate_cairo_code(self, *args: tuple) -> list[str]:
        """Generate cairo code that performs this operation.

        Parameters:
          - 'n_args' stack values starting with the top ones;
          - an output variable name if 'has_output' is True;

        Operation results should be bound to the output variable
        names.

        """



from __future__ import annotations
from collections.abc import Callable

from transpiler.Operations.EnforcedStack import EnforcedStack


class Binary(EnforcedStack):
    def __init__(self):
        super().__init__(n_args=2, has_output=True)


class SimpleBinary(Binary):
    def __init__(
        self,
        eager_eval: Callable[[int, int], int],
        module: str,
        function_name: str,
    ):
        super().__init__()
        self.eager_eval = eager_eval
        self.module = module
        self.function_name = function_name

    def evaluate_eagerly(self, a, b):
        return self.eager_eval(a, b)

    def generate_cairo_code(self, op1, op2, res):
        return [
            f"let (local {res} : Uint256) = {self.function_name}({op1}, {op2})",
        ]

    def required_imports(self):
        return {self.module: {self.function_name}}



from transpiler.Operation import Operation

MAX_VAL = 2 ** 64 - 1

class Gas(Operation):
    @classmethod
    def required_imports(cls):
        return {}

    def proceed(self, state):
        """
        The current approach is to set the gas cost to a big constant,
        instead of doing the gas book keeping in the cairo code.
        It is assumed that the gas cost will not be higher than 64 bits
        """
        state.stack.push_uint256(MAX_VAL)
        return []



from transpiler.Operation import Operation


class CallValue(Operation):
    def proceed(self, state):
        res_ref_name = state.request_fresh_name()
        instruction = self.bind_to_res(res_ref_name)
        state.stack.push_ref(res_ref_name)
        return [instruction]

    def bind_to_res(self, res):
        return f"local {res} : Uint256 = Uint256(1000,1000)"



from transpiler.Operations.Binary import Binary
from transpiler.Operation import Operation
from transpiler.Operations.EnforcedStack import EnforcedStack


class SHA3(EnforcedStack):
    def __init__(self):
        super().__init__(args_spec="ll", n_args=2, has_output=True)

    def generate_cairo_code(self, offset, length, output):
        return [
            f"let (local {output} : Uint256) = sha({offset}, {length})",
            f"local msize = msize",
            "local memory_dict : DictAccess* = memory_dict",
        ]

    def required_imports(self):
        return {"evm.sha3": {"sha"}}


class Keccak256(EnforcedStack):
    def __init__(self):
        super().__init__(args_spec="ll", n_args=2, has_output=True)

    def generate_cairo_code(self, offset, length, output):
        return [
            f"""
                let (local msize) = update_msize(msize, {offset}, {length})
                let (local memval) = create_from_memory({offset}, {length})
                local high
                local low
                %{{
                    from Crypto.Hash import keccak

                    keccak_hash = keccak.new(digest_bits=256)
                    arr = []
                    arr_length = {length}//16 if {length} % 16 == 0 else {length}//16+1
                    for i in range(arr_length):
                        arr.append(memory[ids.memval+i])

                    keccak_input = bytearray()
                    for i in range(arr_length-1):
                        keccak_input += arr[i].to_bytes(16,"big")
                    keccak_input += arr[-1].to_bytes(16,"big") if {length}%16 == 0 else arr[-1].to_bytes({length}%16,"big")

                    keccak_hash.update(keccak_input)
                    hashed = keccak_hash.digest()
                    ids.high = int.from_bytes(hashed[:16],"big")
                    ids.low = int.from_bytes(hashed[16:32],"big")
                %}}
                local {output} : Uint256 = Uint256(low,high)
                """
        ]

    def required_imports(self):
        return {"evm.utils": {"update_msize"}, "evm.array": {"create_from_memory"}}



from __future__ import annotations
from dataclasses import dataclass
from eth_hash.auto import keccak
from enum import Enum
import abc

class Language(Enum):
    SOL = 0
    VYPER = 1

class Contract(abc.ABC):
    @abc.abstractmethod
    def get_compiler_dir(self):
        pass

    @abc.abstractmethod
    def get_code(self, file_path):
        pass

    @abc.abstractmethod
    def verify_source_version(self, code):
        pass

    @abc.abstractmethod
    def compile_contract(self, source_path):
        pass

    @abc.abstractmethod
    def get_bytecode(self, compile_obj):
        pass

    @abc.abstractmethod
    def get_abi(self, path):
        pass




from transpiler.Operation import Operation, NoParse
from transpiler.Operations.EnforcedStack import EnforcedStack


class Log(EnforcedStack):
    def __init__(self, topics_amount: int):
        if not (0 <= topics_amount <= 4):
            raise ValueError(f"LOG can take from 0 to 4 topics, not {topics_amount}")
        super().__init__(args_spec="ll" + "w" * topics_amount, has_output=False)

    @classmethod
    def parse_from_words(cls, words, pos):
        if not words[pos].startswith("LOG"):
            raise NoParse(pos)
        topics_amount = int(words[pos][3:])
        return (cls(topics_amount), pos + 1)

    @classmethod
    def associated_words(cls):
        return [f"LOG{i}" for i in range(5)]

    def generate_cairo_code(self, offset, length, *topics):
        """
        Log operation is currently implemented as no-op.
        It has the proper behavior in regards to popping things from stack
        and updating the memory consumption counter
        """
        return [
            f"let (local msize) = update_msize(msize, {offset}, {length})",
            "local memory_dict : DictAccess* = memory_dict",
        ]

    def required_imports(self):
        return {"evm.utils": {"update_msize"}}



from __future__ import annotations
import abc

from transpiler.Imports import Imports


class Operation(abc.ABC):
    @classmethod
    def parse_from_words(cls, words: list[str], pos: int) -> tuple(cls, int):
        """Given a list of words and a current position, try to parse the
        operation as a word.

        Return `(op, pos')`, where `op` is the parsed operation and
        `pos'` is a new position in the list after parsing.

        Throw NoParse if parsing is unsuccessful.

        """
        word = cls.__name__.upper()
        if words[pos] != word:
            raise NoParse(pos)
        return (cls(), pos + 1)

    def size_in_bytes(self):
        """Size of the operation in bytes. Might be different from 1 for PUSH
        operations, which take argument of variable size.
        """
        return 1

    @classmethod
    def associated_words(cls) -> list[str]:
        """E.g. [POP], [MLOAD], [PUSH1, .., PUSH32]"""
        return [cls.__name__.upper()]

    @classmethod
    def required_imports(cls) -> Imports:
        """Returns a dictionary of imports required to use this operation. A
        dictionary key is a module to import from and a dictionary
        value is a set of names from that module to import.

        """
        return {}

    @classmethod
    def inspect_program(self, code):
        pass

    @abc.abstractmethod
    def proceed(self, state: "EvmToCairo.SegmentState") -> list[str]:
        pass

    def process_structural_changes(self, evmToCairo: "EvmToCairo"):
        """Perform additional transformations on the transpilation process,
        after all state changes and cairo instructions have been
        processed.

        Primarily a hack to allow 'JUMPDEST' to create new segments.

        """


class NoParse(BaseException):
    def __init__(self, word_no: int):
        self.word_no = word_no


# The following piece of code serves for automatic discovery of
# operations.
import importlib
import os

with os.scandir("warp/transpiler/Operations") as it:
    for entry in it:
        if entry.is_file() and entry.name.endswith(".py"):
            importlib.import_module(f".{entry.name[:-3]}", package="transpiler.Operations")



from __future__ import annotations
import os, sys
from typing import Dict, List, Tuple

WARP_ROOT = os.path.abspath(os.path.join(__file__, "../../../.."))
sys.path.append(os.path.join(WARP_ROOT, "src"))
import os, sys
from cli.compilation.utils import is_entry_seq, get_jumpdest_offset
from cli.compilation.opcode_tables import OP_TO_STR
import json


class InstructionIterator:
    def __init__(self, bytecode):
        self.bytecode = [bytecode[x : x + 2] for x in range(0, len(bytecode), 2)]
        self.op_table = OP_TO_STR
        self.cur_op = self.bytecode[0]
        self.cur_op_str = self.op_str()
        self.arg = []
        self.pc = 0
        self.started = False
        self.error = None

    def next(self) -> bool:
        if len(self.bytecode) <= self.pc or self.error != None:
            return False
        if self.started:
            if self.arg != None:
                self.pc += len(self.arg)
            self.pc += 1
        else:
            self.started = True

        if len(self.bytecode) <= self.pc:
            return False

        self.cur_op = self.bytecode[self.pc]
        self.cur_op_str = self.op_str()
        push_op, size = self.is_push(self.cur_op_str)
        if push_op:
            u = self.pc + size + 1
            if len(self.bytecode) <= self.pc or len(self.bytecode) < u:
                self.error = f"Incomplete push instruction at {self.pc}\n curr_op: {self.cur_op_str}"
            self.arg = self.bytecode[self.pc + 1 : u]
            self.arg_padded = self.pad_arg(self.arg)
        else:
            self.arg = None
        return True

    def disassemble(self) -> Dict[str, str]:
        self.instructions = {}
        while self.next():
            if self.arg != None and len(self.arg) > 0:
                self.instructions[hex(self.pc)] = f"{self.op_str()} {self.arg_padded}"
            else:
                self.instructions[hex(self.pc)] = f"{self.op_str()}"
        return self.instructions

    def is_push(self, op_str) -> Tuple[bool, int]:
        if "PUSH" in op_str:
            return True, self.push_arg_size(op_str)
        else:
            return False, 0

    def create_entrypoint(self, instructions: List[str], selectors, lang):
        selector = instructions[1]
        offset = get_jumpdest_offset(lang)
        jumpdest = instructions[offset]
        func_sig = selectors[instructions[1]]
        return {
            "signature": func_sig,
            "selector": selector,
            "jumpdest": jumpdest,
        }

    def push_arg_size(self, op_str) -> int:
        return int(op_str[op_str.find("H") + 1 :])

    def op_str(self):
        try:
            return self.op_table[self.cur_op]
        except KeyError:
            val = f"INVALID"
            return val

    def pad_arg(self, arg):
        len_push = self.push_arg_size(self.cur_op_str) * 2
        arg_hex = hex(int("".join(arg), 16))
        l_arg = len(arg_hex[2:])
        arg = "0x" + "0" * (len_push - l_arg) + arg_hex[2:]
        return arg


if __name__ == "__main__":
    with open("weth10.bin") as f:
        bytecode = f.read()
    it = InstructionIterator(bytecode)
    dis = it.disassemble()
    with open("weth.json", "w") as f:
        json.dump(dis, f, indent=4)



import abc



def is_valid_uintN(n: int, x: int):
    return 0 <= x < 2 ** n


UINT256_BOUND = 2 ** 256
UINT256_HALF_BOUND = 2 ** 255
UINT128_BOUND = 2 ** 128


class StackValue:
    @abc.abstractmethod
    def __str__(self):
        pass

    @abc.abstractmethod
    def get_low_high(self):
        pass

    def get_low_bits(self):
        return self.get_low_high()[0]


# SHOULD BE IMMUTABLE
class Uint256(StackValue):
    def __init__(self, x: int):
        if not is_valid_uintN(256, x):
            raise ValueError(f"{x} is not a valid uint256")
        self.x = x

    def __str__(self):
        high, low = divmod(self.x, UINT128_BOUND)
        return f"Uint256({low}, {high})"

    def __add__(self, y: int):
        return Uint256((self.x + y) % UINT256_BOUND)

    def __sub__(self, y: int):
        return Uint256((self.x - y) % UINT256_BOUND)

    def get_int_repr(self) -> str:
        return repr(self.x)

    def get_low_high(self):
        high, low = divmod(self.x, UINT128_BOUND)
        return low, high


class Str(StackValue):
    def __init__(self, x: str):
        self.x = x

    def __str__(self):
        return self.x

    def get_low_high(self):
        return f"{self.x}.low", f"{self.x}.high"

    def get_low_bits(self):
        return f"{self.x}.low"


from transpiler.Operation import Operation, NoParse
from transpiler.utils import is_valid_uintN


class Push(Operation):
    def __init__(self, bytes_in_value: int, value: int):
        if not (1 <= bytes_in_value <= 32):
            raise ValueError(
                f"PUSH can push values from 1 to 32 bytes, not {bytes_in_value}"
            )
        bits = 8 * bytes_in_value
        if not is_valid_uintN(bits, value):
            raise ValueError(f"{value} is not a valid uint{bits}")
        self.bytes_in_value = bytes_in_value
        self.value = value

    @classmethod
    def parse_from_words(cls, words, pos):
        if not words[pos].startswith("PUSH"):
            raise NoParse(pos)
        bytes_in_value = int(words[pos][4:])
        value = int(words[pos + 1], 16)
        return (cls(bytes_in_value, value), pos + 2)

    @classmethod
    def associated_words(cls):
        return [f"PUSH{i}" for i in range(1, 33)]

    def size_in_bytes(self):
        return 1 + self.bytes_in_value

    def proceed(self, state):
        state.stack.push_uint256(self.value)
        return []


class Dup(Operation):
    def __init__(self, depth: int):
        if not (1 <= depth <= 16):
            raise ValueError(f"DUP depth must be in between 1 and 16, not {depth}")
        self.depth = depth

    @classmethod
    def parse_from_words(cls, words, pos):
        if not words[pos].startswith("DUP"):
            raise NoParse(pos)
        depth = int(words[pos][3:])
        return (cls(depth), pos + 1)

    @classmethod
    def associated_words(cls):
        return [f"DUP{i}" for i in range(1, 17)]

    def proceed(self, state):
        state.stack.dup(self.depth)
        return []


class Swap(Operation):
    def __init__(self, depth: int):
        if not (1 <= depth <= 16):
            raise ValueError(f"SWAP depth must be in between 1 and 16, not {depth}")
        self.depth = depth

    @classmethod
    def parse_from_words(cls, words, pos):
        if not words[pos].startswith("SWAP"):
            raise NoParse(pos)
        depth = int(words[pos][4:])
        return (cls(depth), pos + 1)

    @classmethod
    def associated_words(cls):
        return [f"SWAP{i}" for i in range(1, 17)]

    def proceed(self, state):
        state.stack.swap(self.depth)
        return []


class Pop(Operation):
    def proceed(self, state):
        state.stack.pop()
        return []



from __future__ import annotations
from collections import OrderedDict
import os, sys
import json
import subprocess
from typing import List
from semantic_version import Version
from vyper.cli.vyper_compile import compile_files
from web3 import Web3

WARP_ROOT = os.path.abspath(os.path.join(__file__, "../../../.."))
sys.path.append(os.path.join(WARP_ROOT, "src"))
from cli.compilation.utils import (
    get_selectors,
    get_selector_jumpdests,
    is_entry_seq,
    is_payable_check_seq,
)
from cli.compilation.Disasm import InstructionIterator
from cli.compilation.Contract import Contract, Language
import solcx


class Vyper(Contract):
    def __init__(self, source_path):
        self.w3 = Web3()
        self.lang = Language.VYPER
        self.compiler_path = self.get_compiler_dir()
        self.source_path = source_path
        self.source_name = os.path.basename(source_path)
        self.source_base_dir = os.path.dirname(self.source_path)
        self.version = self.get_compiler_ver(self.compiler_path)
        self.code = self.get_code(self.source_path)
        self.bytecode = self.get_bytecode(self.source_path)
        self.abi = self.get_abi(self.source_path)
        self.selectors = get_selectors(self.abi)
        self.opcodes = self.get_opcodes()
        self.opcodes_str = self.get_opcode_str(self.opcodes)
        self.selector_jumpdests = get_selector_jumpdests(self)
        self.web3_interface = self.w3.eth.contract(abi=self.abi, bytecode=self.bytecode)

    def get_compiler_dir(self) -> str:
        return (
            subprocess.run(["which", "vyper"], stdout=subprocess.PIPE)
            .stdout.decode()
            .strip()
        )

    def get_compiler_ver(self, vyper_path: str) -> str:
        vy_ver = (
            subprocess.run(
                [sys.executable, vyper_path, "--version"], stdout=subprocess.PIPE
            )
            .stdout.decode()
            .strip()
        )
        return vy_ver[: vy_ver.find("+")]

    def install_compiler_version(self, version: str):
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install", f"vyper=={version}"]
        )
        self.version = version

    def get_code(self, source_path: str) -> List[str]:
        with open(source_path) as f:
            code = f.read()
            f.seek(0)
            code_split = "\n".join(
                [line.strip() for line in f.read().split("\n") if line.strip()]
            ).split("\n")
        return code_split

    def compile_contract(self, source_path: str) -> OrderedDict:
        self.verify_source_version(self.code)
        pwd = os.getcwd()
        base_source_dir = os.path.abspath(os.path.join(source_path, "../"))
        os.chdir(base_source_dir)
        compiled = compile_files([source_path], ["bytecode_runtime"])
        os.chdir(pwd)
        return compiled

    def get_bytecode(self, source_path: str) -> str:
        base_source_dir = os.path.abspath(os.path.join(source_path, "../"))
        if not os.path.isdir(self.artifacts_dir):
            os.makedirs(self.artifacts_dir)
        try:
            compiled = self.compile_contract(source_path)
            return compiled[self.source_name]["bytecode_runtime"][2:]
        except:
            compiler_dir = self.get_compiler_dir()
            res = (
                subprocess.run(
                    [
                        sys.executable,
                        f"{compiler_dir}",
                        "-f",
                        "bytecode_runtime",
                        source_path,
                    ],
                    stdout=subprocess.PIPE,
                )
                .stdout.decode()
                .strip()
            )
            return res[2:]

    def verify_source_version(self, code_split: List[str]):
        source_ver = ""
        for line in code_split:
            if "@version" in line:
                source_ver = line
        if source_ver == "":
            raise Exception("No source version specified in contract")
        source_ver = source_ver[source_ver.find("0") :].strip()
        if source_ver != self.version:
            self.install_compiler_version(source_ver)
            ver = self.get_compiler_ver(self.get_compiler_dir())
        return

    def get_abi(self, source_path):
        self.verify_source_version(self.code)
        pwd = os.getcwd()
        base_source_dir = os.path.abspath(os.path.join(source_path, "../"))
        os.chdir(base_source_dir)
        try:
            compiled = compile_files([source_path], ["abi"])
            return compiled
        except:
            compiler_dir = self.get_compiler_dir()
            res = (
                subprocess.run(
                    [sys.executable, f"{compiler_dir}", "-f", "abi", source_path],
                    stdout=subprocess.PIPE,
                )
                .stdout.decode()
                .strip()
            )
            return json.loads(res)
        os.chdir(pwd)
        return self.compiled[self.source_name]["abi"][2:]

    def get_opcodes(self):
        it = InstructionIterator(self.bytecode)
        return " ".join(x for x in it.disassemble().values()).split(" ")

    def get_opcode_str(self, opcodes):
        return " ".join(x for x in opcodes)


class Solidity(Contract):
    def __init__(self, source_path):
        w3 = Web3()
        self.lang = Language.SOL
        self.compiler_path = self.get_compiler_dir()
        self.source_path = source_path
        self.source_name = os.path.basename(source_path)
        self.source_base_dir = os.path.dirname(self.source_path)
        self.artifacts_dir = os.path.join(os.path.expanduser("~"), ".warp", "artifacts")
        self.code = self.get_code(source_path)
        self.compiled, self.version = self.compile_contract(self.code)
        self.bytecode = self.get_bytecode(self.compiled, self.version)
        self.abi = self.get_abi(self.code)
        self.opcodes = self.get_opcodes()
        self.selectors = get_selectors(self.abi)
        self.selector_jumpdests = get_selector_jumpdests(self)
        self.web3_interface = w3.eth.contract(abi=self.abi, bytecode=self.bytecode)
        self.opcodes_str = self.get_opcode_str(self.opcodes)

    def get_opcode_str(self, opcodes):
        return " ".join(x for x in opcodes)

    def get_compiler_dir(self):
        return (
            subprocess.run(["which", "solc"], stdout=subprocess.PIPE)
            .stdout.decode()
            .strip()
        )

    def get_source_version(self, code):
        code_split = code.split("\n")
        for line in code_split:
            if "pragma" in line:
                return line[line.index("0.") :].replace(";", "")
        raise Exception("No Solidity version specified in contract")

    def get_code(self, file_path):
        with open(file_path) as f:
            code = f.read()
        return code

    def compile_contract(self, code):
        self.verify_source_version(code)
        pwd = os.getcwd()
        base_source_dir = os.path.abspath(os.path.join(self.source_path, "../"))
        if not os.path.isdir(self.artifacts_dir):
            os.makedirs(self.artifacts_dir)
        os.chdir(base_source_dir)
        source_version = solcx.set_solc_version_pragma(code)
        if source_version.minor >= 6:
            compiled = solcx.compile_source(
                code,
                metadata_hash="none",
                output_values=["bin-runtime"],
                solc_version=source_version,
            )
        else:
            compiled = solcx.compile_source(
                code,
                output_values=["bin-runtime"],
                solc_version=source_version,
            )
        os.chdir(pwd)
        return compiled, source_version

    def get_abi(self, code):
        self.verify_source_version(code)
        pwd = os.getcwd()
        base_source_dir = os.path.abspath(os.path.join(self.source_path, "../"))
        os.chdir(base_source_dir)
        abi = solcx.compile_source(code, output_values=["abi"])
        abi_succinct = []
        count = 0
        for k, v in abi.items():
            new_k = k.split(":")[1]
            if new_k.startswith("I"):
                continue
            elif v["abi"] != "":
                abi_succinct = v["abi"]
                count += 1
            if count > 1:
                raise Exception("There were more than one non-interface abis")
        os.chdir(pwd)
        with open(
            os.path.join(self.artifacts_dir, f"{self.source_name[:-4]}_abi.json"),
            "w",
        ) as f:
            json.dump(abi_succinct, f, indent=4)
        return abi_succinct

    def verify_source_version(self, code):
        source_version = Version(self.get_source_version(code))
        installed_vers = solcx.get_installed_solc_versions()
        if source_version not in installed_vers:
            solcx.install_solc_pragma(code)
            solcx.set_solc_version_pragma(code)
        return source_version

    # only one of the keys should have their bin-runtime value
    # non-empty.
    def get_bytecode(self, compile_obj, ver):
        bytecode = ""
        for k, v in compile_obj.items():
            if v["bin-runtime"] == "":
                continue
            else:
                bytecode += v["bin-runtime"]
        # remove the insane metadata hash
        if ver.minor <= 6:
            try:
                idx = bytecode.index("a165")
                metadata = bytecode[idx:]
                if "5820" in metadata and metadata.endswith("0029"):
                    bytecode = bytecode[:idx]
            except:
                idx = bytecode.index("a265")
                metadata = bytecode[idx:]
                if "5820" in metadata and metadata.endswith("0032"):
                    bytecode = bytecode[:idx]

        with open(
            os.path.join(self.artifacts_dir, f"{self.source_name[:-4]}_bytecode"),
            "w",
        ) as f:
            f.write(bytecode)
        return bytecode

    def get_opcodes(self):
        it = InstructionIterator(self.bytecode)
        return " ".join(x for x in it.disassemble().values()).split(" ")



from transpiler.Operations.EnforcedStack import EnforcedStack


class CallDataCopy(EnforcedStack):
    def __init__(self):
        super().__init__(args_spec="lll", has_output=False)

    def generate_cairo_code(self, dest_offset, offset, length):
        return [
            f"let (local msize) = update_msize(msize, {dest_offset}, {length})",
            "local memory_dict : DictAccess* = memory_dict",
            "copy_to_memory(exec_env.input_len, exec_env.input, ",
            f"   {dest_offset}, {offset}, {length})",
            "local memory_dict : DictAccess* = memory_dict",
        ]

    @classmethod
    def required_imports(cls):
        return {
            "evm.array": {"copy_to_memory"},
            "evm.utils": {"update_msize"},
        }


class CallDataSize(EnforcedStack):
    def __init__(self):
        super().__init__(n_args=0, has_output=True)

    def generate_cairo_code(self, res):
        return [f"local {res} : Uint256 = Uint256(exec_env.calldata_size, 0)"]


class CallDataLoad(EnforcedStack):
    def __init__(self):
        super().__init__(args_spec="l", has_output=True)

    def generate_cairo_code(self, byte_pos, res):
        return [
            f"let (local {res} : Uint256) = aload(exec_env.input_len, exec_env.input, {byte_pos})",
        ]

    @classmethod
    def required_imports(cls):
        return {"evm.array": {"aload"}}



from __future__ import annotations
import os
from utils import from_abi, read_file_json, get_selectors

WARP_ROOT = os.path.abspath(os.path.join(__file__, "../../.."))

TEST_CONTRACTS = ["link", "weth", "weth10"]


def test_get_selectors(contracts: List[str] = TEST_CONTRACTS):
    for abi in contracts:
        abi_path = os.path.join(WARP_ROOT, "tests", "utils", f"{abi}.abi")
        func_selectors = get_selectors(abi_path)
        opcodes_path = os.path.join(WARP_ROOT, "tests", "utils", f"{abi}.opcode")
        with open(opcodes_path) as f:
            opcodes = f.read()
        opcodes = opcodes.lower()
        for k in func_selectors.keys():
            # solidity and vyper remove leading zeros when you use the
            # --opcodes flag to ouptut the opcodes from the compiled contract.
            if "0" in k[2:]:
                k = "0x" + k[2:].lstrip("0")
            assert k in opcodes, f"failed with {k} in file {abi}"
            if k not in opcodes:
                return 0
    return 1



from setuptools import setup, find_packages
from io import open
from os import path
import pathlib

# The directory containing this file
HERE = pathlib.Path(__file__).parent
# The text of the README file
with open(path.join(HERE, "README.md")) as f:
    README = f.read()

# automatically captured required modules for install_requires in requirements.txt and as well as configure dependency links
with open(path.join(HERE, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')
install_requires = [x.strip() for x in all_reqs if ('git+' not in x) and (
    not x.startswith('#')) and (not x.startswith('-'))]
dependency_links = [x.strip().replace('git+', '') for x in all_reqs \
                    if 'git+' not in x]

setup (
 name = 'evm-warp',
 description='Transpile EVM-Compatible Languages To Cairo',
 version='0.1.0',
 package_dir={"": "warp"},
 packages=["cli","cli.compilation", "transpiler", "transpiler.Operations"], # list of all packages
 include_package_data=True,
 package_data={'': ['*.json']},
 install_requires=install_requires,
 python_requires='>=3.7', # any python greater than 3.7
 entry_points='''
        [console_scripts]
        warp=cli.warp_cli:main
    ''',
 author="Nethermind",
 keyword="Ethereum, Layer2, ETH, StarkNet, Nethermind, StarkWare, transpilation, warp, transpiler, cairo",
 long_description=README,
 long_description_content_type="text/markdown",
 license='Apache 2.0',
 url='https://github.com/NethermindEth/warp',
 download_url='',
  dependency_links=dependency_links,
  author_email='vardygreg23@gmail.com',
  classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ]
)


from transpiler.Operation import Operation
from transpiler.StackValue import Uint256
from transpiler.Operations.EnforcedStack import EnforcedStack
from transpiler.Operations.Unary import Unary
from transpiler.utils import get_low_high

# The view functions get_storage_low/high will are used
# for starknet testing.
STORAGE_DECLS = """
@storage_var
func evm_storage(low: felt, high: felt, part: felt) -> (res : felt):
end

func s_load{
        storage_ptr: Storage*, range_check_ptr, pedersen_ptr: HashBuiltin*}(
        low: felt, high: felt) -> (res : Uint256):
    let (low_r) = evm_storage.read(low, high, 1)
    let (high_r) = evm_storage.read(low, high, 2)
    return (Uint256(low_r, high_r))
end

func s_store{
        storage_ptr: Storage*, range_check_ptr, pedersen_ptr: HashBuiltin*}(
        low: felt, high: felt, value_low : felt, value_high : felt):
    evm_storage.write(low=low, high=high, part=1, value=value_low)
    evm_storage.write(low=low, high=high, part=2, value=value_high)
    return ()
end

@view
func get_storage_low{
        storage_ptr : Storage*, range_check_ptr, pedersen_ptr : HashBuiltin*}(
        low : felt, high : felt) -> (res : felt):
    let (storage_val_low) = evm_storage.read(low=low, high=high, part=1)
    return (res=storage_val_low)
end

@view
func get_storage_high{
        storage_ptr : Storage*, range_check_ptr, pedersen_ptr : HashBuiltin*}(
        low : felt, high : felt) -> (res : felt):
    let (storage_val_high) = evm_storage.read(low=low, high=high, part=2)
    return (res=storage_val_high)
end
"""


class SStore(EnforcedStack):
    def __init__(self):
        super().__init__(n_args=2, has_output=False)

    def generate_cairo_code(self, loc, value):
        loc_low, loc_high = get_low_high(loc)
        value_low, value_high = get_low_high(value)
        return [
            f"s_store(low={loc_low}, high={loc_high}, "
            f"value_low={value_low}, value_high={value_high})",
            "local pedersen_ptr : HashBuiltin* = pedersen_ptr",
            "local storage_ptr : Storage* = storage_ptr",
        ]

    def process_structural_changes(self, evmToCairo):
        evmToCairo.requires_storage = True


class SLoad(Unary):
    # return zero'd values if we are trying to load from an
    # uninitialized storage slot
    def generate_cairo_code(self, loc, res):
        loc_low, loc_high = get_low_high(loc)
        return [
            f"let (local {res} : Uint256) = s_load(low={loc_low}, high={loc_high})",
            "local pedersen_ptr : HashBuiltin* = pedersen_ptr",
            "local storage_ptr : Storage* = storage_ptr",
            ]

    def process_structural_changes(self, evmToCairo):
        evmToCairo.requires_storage = True



from __future__ import annotations
import sys
from dataclasses import dataclass
from collections import defaultdict

from starkware.cairo.lang.compiler.parser import parse_file

from transpiler.Operation import Operation
from transpiler.Imports import UINT256_MODULE, format_imports, merge_imports
from transpiler.EvmStack import EvmStack
import transpiler.StackValue as StackValue
from transpiler.Operations.Storage import STORAGE_DECLS

LANGUAGE = "%lang starknet"
BUILTINS = ["pedersen", "range_check"]

COMMON_IMPORTS = {
    "starkware.cairo.common.registers": {"get_fp_and_pc"},
    "starkware.cairo.common.dict_access": {"DictAccess"},
    "starkware.cairo.common.math_cmp": {"is_le"},
    "starkware.cairo.common.default_dict": {
        "default_dict_new",
    },
    UINT256_MODULE: {"Uint256", "uint256_eq"},
    "starkware.cairo.common.cairo_builtins": {"HashBuiltin"},
    "starkware.starknet.common.storage": {"Storage"},
    "evm.stack": {"StackItem"},
    "evm.output": {"Output"},
    "evm.exec_env": {"ExecutionEnvironment"},
}

MAIN = """
@external
func main{storage_ptr: Storage*, pedersen_ptr: HashBuiltin*, range_check_ptr}(
   calldata_size, unused_bytes, input_len : felt, input: felt*
   ):
   alloc_locals
   let (local __fp__, _) = get_fp_and_pc()
   local exec_env: ExecutionEnvironment = ExecutionEnvironment(
      calldata_size = calldata_size,
      input_len = input_len,
      input = input,
   )

   let (local memory_dict : DictAccess*) = default_dict_new(0)
   local memory_start : DictAccess* = memory_dict
   
   tempvar msize = 0

   local stack0 : StackItem
   assert stack0 = StackItem(value=Uint256(-1, 0), next=&stack0)  # Points to itself.

   let (local stack, local output) = run_from{
      storage_ptr=storage_ptr,
      pedersen_ptr=pedersen_ptr,
      range_check_ptr=range_check_ptr,
      msize=msize, memory_dict=memory_dict
      }(&exec_env, Uint256(0, 0), &stack0)

   return ()
end
"""

NO_OUTPUT = "Output(0, cast(0, felt*), 0)"
EMPTY_OUTPUT = "Output(1, cast(0, felt*), 0)"


@dataclass
class SegmentState:
    stack: EvmStack
    n_locals: int
    unreachable: bool
    msize: int
    cur_evm_pc: StackValue.Uint256

    def __init__(self, cur_evm_pc: StackValue.Uint256):
        self.stack = EvmStack()
        self.n_locals = 0
        self.unreachable = False
        self.msize = 0
        self.cur_evm_pc = cur_evm_pc

    def make_return_instructions(
        self, pc: StackValue, output: str = NO_OUTPUT
    ) -> list[str]:
        build_instructions, stack_ref = self.stack.build_stack_instructions()
        return [
            *build_instructions,
            f"return (stack={stack_ref}, evm_pc={pc}, output={output})",
        ]

    def request_fresh_name(self) -> str:
        var_name = f"tmp{self.n_locals}"
        self.n_locals += 1
        return var_name


class EvmToCairo:
    def __init__(self, cur_evm_pc: int):
        self.code_segments: list[StackValue.Uint256] = []
        self.text_segments = []
        self.state = SegmentState(StackValue.Uint256(cur_evm_pc))
        self.imports = defaultdict(set)
        merge_imports(self.imports, COMMON_IMPORTS)
        self.requires_storage = False

        self.start_new_segment()

    def finish_segment(self):
        int_pc = int(self.segment_pc.get_int_repr())
        if int_pc != 0:
            segment_pc_txt = int_pc - 1
        else:
            segment_pc_txt = self.segment_pc.get_int_repr()
        self.text_segments.append(
            "\n".join(
                [
                    f"func segment{segment_pc_txt}{self.__make_implicit_args()}"
                    "(exec_env: ExecutionEnvironment*, stack : StackItem*) -> "
                    "(stack : StackItem*, evm_pc : Uint256, output: Output):",
                    "alloc_locals",
                    "let stack0 = stack",
                    "let (local __fp__, _) = get_fp_and_pc()",
                ]
                + self.state.stack.prepare()
                + self.instructions
                + ["end"]
            )
        )

    def start_new_segment(self):
        self.instructions = []
        self.state = SegmentState(self.state.cur_evm_pc)
        self.segment_pc = self.state.cur_evm_pc
        self.code_segments.append(self.segment_pc)

    def process_operation(self, operation: Operation):
        self.state.cur_evm_pc += operation.size_in_bytes()
        if not self.state.unreachable:
            more_instructions = operation.proceed(self.state)
            self.instructions.extend(more_instructions)
            merge_imports(self.imports, operation.required_imports())
        operation.process_structural_changes(self)

    def __make_implicit_args(self):
        return "{storage_ptr : Storage*, pedersen_ptr : HashBuiltin*, range_check_ptr, msize, memory_dict: DictAccess*}"

    def construct_run_from_function(self):
        run_from = [
            f"func run_from{self.__make_implicit_args()}"
            "(exec_env: ExecutionEnvironment*, evm_pc: Uint256, stack: StackItem*) "
            "-> (stack: StackItem*, output: Output):"
        ]

        for segment_pc in self.code_segments:

            int_pc = int(segment_pc.get_int_repr())
            if int_pc != 0:
                segment_pc_txt = int_pc - 1
            else:
                segment_pc_txt = int_pc
            run_from.extend(
                [
                    f"let (immediate) = uint256_eq{{range_check_ptr=range_check_ptr}}(evm_pc, Uint256({segment_pc_txt}, 0))",
                    f"if immediate == 1:",
                    f"let (stack, evm_pc, output) = segment{segment_pc_txt}(exec_env, stack)",
                    "if output.active == 1:",
                    "return (stack, output)",
                    "end",
                    f"return run_from(exec_env, evm_pc, stack)",
                    "end",
                ]
            )

        run_from.extend(
            [
                f"let (immediate) = uint256_eq{{range_check_ptr=range_check_ptr}}(evm_pc, Uint256(-1, 0))",
                f"if immediate == 1:",
                f"return (stack, {EMPTY_OUTPUT})",
                "end",
            ]
        )

        run_from.append("# Fail.")
        run_from.append("assert 0 = 1")
        run_from.append("jmp rel 0")
        run_from.append("end")
        return "\n".join(run_from)

    def finish(self, dump_all):
        if dump_all and not self.state.unreachable:
            insts, new_stack = self.state.stack.build_stack_instructions()
            self.instructions.extend(insts)
            self.instructions.append(
                f"return (stack={new_stack}, "
                f"evm_pc=Uint256(0, 0), output={EMPTY_OUTPUT})"
            )
        self.finish_segment()
        builtins_line = "%builtins " + " ".join(BUILTINS) if BUILTINS else ""
        return "\n\n".join(
            [
                LANGUAGE,
                builtins_line,
                format_imports(self.imports),
            ]
            + ([STORAGE_DECLS] if self.requires_storage else [])
            + [
                *self.text_segments,
                self.construct_run_from_function(),
                self.__make_main(),
            ]
        )

    def __make_main(self):
        return MAIN


def get_subclasses(cls):
    direct = cls.__subclasses__()
    indirect = [subsubcls for subcls in direct for subsubcls in get_subclasses(subcls)]
    return direct + indirect


def read_words(file):
    for line in file:
        yield from line.partition("#")[0].split()  # skip comments


def parse_operations(file):
    words = tuple(read_words(file))
    operation_classes = get_subclasses(Operation)

    word_to_cls = {}
    for cls in operation_classes:
        word_to_cls.update({word: cls for word in cls.associated_words()})
    i = 0
    while i < len(words):
        correct_cls = word_to_cls[words[i]]
        (op, i) = correct_cls.parse_from_words(words, i)
        yield op

def parse_ops_direct(words):
    operation_classes = get_subclasses(Operation)

    word_to_cls = {}
    for cls in operation_classes:
        word_to_cls.update({word: cls for word in cls.associated_words()})

    i = 0
    while i < len(words):
        correct_cls = word_to_cls[words[i]]
        (op, i) = correct_cls.parse_from_words(words, i)
        yield op

USAGE = f"Usage: python {sys.argv[0]} EVM-OPCODES-FILE [--dump-all]"


def main():
    argc = len(sys.argv)
    if argc == 2:
        dump_all = False
    elif argc == 3:
        if sys.argv[2] != "--dump-all":
            sys.exit(USAGE)
        dump_all = True
    else:
        sys.exit(USAGE)

    evm_to_cairo = EvmToCairo(cur_evm_pc=0)

    with open(sys.argv[1], "r") as opcodes_file:
        operations = list(parse_operations(opcodes_file))
        for op in operations:
            op.inspect_program(operations)

        for op in operations:
            evm_to_cairo.process_operation(op)

    print(parse_file(evm_to_cairo.finish(dump_all)).format())


if __name__ == "__main__":
    main()



from transpiler.Operation import Operation

class PC(Operation):
    @classmethod
    def required_imports(cls):
        return {}

    def proceed(self, state):
        state.stack.push_uint256(state.cur_evm_pc - 1)
        return []



import operator
from transpiler.Operations.Binary import Binary, SimpleBinary
from transpiler.Operations.Unary import Unary
from transpiler.Imports import UINT256_MODULE
from transpiler.utils import uint256_to_int256, int256_to_uint256


class And(SimpleBinary):
    def __init__(self):
        super().__init__(operator.and_, UINT256_MODULE, "uint256_and")


class Or(SimpleBinary):
    def __init__(self):
        super().__init__(operator.or_, UINT256_MODULE, "uint256_or")


class Not(Unary):
    def evaluate_eagerly(self, a):
        return ~a

    def generate_cairo_code(self, op1, res):
        return [f"let (local {res} : Uint256) = uint256_not({op1})"]

    @classmethod
    def required_imports(cls):
        return {UINT256_MODULE: {"uint256_not"}}


class Xor(SimpleBinary):
    def __init__(self):
        super().__init__(operator.xor, UINT256_MODULE, "uint256_xor")


class Shl(SimpleBinary):
    def __init__(self):
        super().__init__(operator.lshift, UINT256_MODULE, "uint256_shl")


class Shr(SimpleBinary):
    def __init__(self):
        super().__init__(operator.rshift, UINT256_MODULE, "uint256_shr")


def sar(a, b):
    b = uint256_to_int256(b)  # per yellow paper, only b is treated as signed
    return int256_to_uint256(a >> b)


class Sar(SimpleBinary):
    def __init__(self):
        super().__init__(sar, UINT256_MODULE, "uint256_sar")


class Byte(Binary):
    def evaluate_eagerly(self, b, x):
        return (x >> (248 - b * 8)) & 0xFF

    def generate_cairo_code(self, op1, op2, res):
        # the reverse order in `uint256_byte` call is intentional
        return [f"let (local {res} : Uint256) = uint256_byte({op2}, {op1})"]

    def required_imports(cls):
        return {UINT256_MODULE: {"uint256_byte"}}



from transpiler.Imports import UINT256_MODULE
from transpiler.Operations.Binary import Binary
from transpiler.Operations.Unary import Unary


class IsZero(Unary):
    def bind_to_res(self, operand, res, evaluatable):
        if not evaluatable:
            return f"let (local {res} : Uint256) = is_zero({operand})", False, 0
        else:
            evaluated = 1 if operand == 0 else 0
            return (
                "",
                True,
                evaluated,
            )

    @classmethod
    def required_imports(cls):
        return {"evm.uint256": {"is_zero"}}


class Eq(Binary):
    def bind_to_res(self, op1, op2, res, evaluatable):
        if not evaluatable:
            return f"let (local {res} : Uint256) = is_eq({op1}, {op2})", False, 0
        else:
            evaluated = 1 if op1 == op2 else 0
            return (
                "",
                True,
                evaluated,
            )

    @classmethod
    def required_imports(cls):
        return {"evm.uint256": {"is_eq"}}


class Lt(Binary):
    def bind_to_res(self, op1, op2, res, evaluatable):
        if not evaluatable:
            return f"let (local {res} : Uint256) = is_lt({op1}, {op2})", False, 0
        else:
            evaluated = 1 if op1 < op2 else 0
            return (
                "",
                True,
                evaluated,
            )

    @classmethod
    def required_imports(cls):
        return {"evm.uint256": {"is_lt"}}


class Gt(Binary):
    def bind_to_res(self, op1, op2, res, evaluatable):
        if not evaluatable:
            return f"let (local {res} : Uint256) = is_gt({op1}, {op2})", False, 0
        else:
            evaluated = 1 if op1 > op2 else 0
            return (
                "",
                True,
                evaluated,
            )

    @classmethod
    def required_imports(cls):
        return {"evm.uint256": {"is_gt"}}


class Slt(Binary):
    def bind_to_res(self, op1, op2, res, evaluatable):
        if not evaluatable:
            return f"let (local {res} : Uint256) = slt({op1}, {op2})", False, 0
        else:
            evaluated = self.slt_eval(op1, op2)
            return (
                "",
                True,
                evaluated,
            )

    def slt_eval(self, op1, op2):
        op1 = -op1 if op1 > (2 ** 255 - 1) else op1
        op2 = -op2 if op1 > (2 ** 255 - 1) else op2
        return 1 if op1 < op2 else 0

    @classmethod
    def required_imports(cls):
        return {"evm.uint256": {"slt"}}


class Sgt(Binary):
    def bind_to_res(self, op1, op2, res, evaluatable):
        if not evaluatable:
            return f"let (local {res} : Uint256) = sgt({op1}, {op2})", False, 0
        else:
            evaluated = self.sgt_eval(op1, op2)
            return (
                "",
                True,
                evaluated,
            )

    def sgt_eval(self, op1, op2):
        op1 = -op1 if op1 > (2 ** 255 - 1) else op1
        op2 = -op2 if op1 > (2 ** 255 - 1) else op2
        return 1 if op1 > op2 else 0

    @classmethod
    def required_imports(cls):
        return {"evm.uint256": {"sgt"}}
